@@ -108,7 +108,7 @@
 
 unsigned int mtools_lock_timeout=30;
 
-void serve_client(int sock, char **device_name, unsigned int n_dev,
+void serve_client(int sock, const char *const*device_name, unsigned int n_dev,
 		  int close_stderr);
 
 
@@ -363,8 +363,9 @@ static char recv_packet(Packet packet, io_buffer fp, Dword maxlength)
 }
 
 static ssize_t read_packet(Packet packet, int fd, Dword length) {
+	ssize_t ret;
 	make_new(packet, length);
-	ssize_t ret = read(fd, packet->data, packet->len);
+	ret = read(fd, packet->data, packet->len);
 	if(ret < 0)
 		return ret;
 	packet->len = (Dword) ret;
@@ -559,14 +560,14 @@ static uint16_t getportnum(char *portnum)
 
 	for (port = 0; isdigit(*digits); ++digits)
 		{
-			port = (port * 10) + (*digits - '0');
+			port = (port * 10) + (uint8_t)(*digits - '0');
 		}
 
 	if ((*digits != '\0') || (port <= 0))
 		{
 			if ((serv = getservbyname(portnum, "tcp")) != NULL)
 				{
-					port = ntohs(serv->s_port);
+					port = ntohs((uint16_t)serv->s_port);
 				}
 			else
 				{
@@ -749,9 +750,9 @@ static void alarm_signal(int a UNUSEDP)
 /*
  * This is the main loop when running as a server.
  */
-static void server_main_loop(int sock, char **device_name,
-			     unsigned  int n_dev) NORETURN;
-static void server_main_loop(int sock, char **device_name,
+static void server_main_loop(int sock, const char *const*device_name,
+			     unsigned int n_dev) NORETURN;
+static void server_main_loop(int sock, const char *const*device_name,
 			     unsigned int n_dev)
 {
 	struct sockaddr_in	addr;
@@ -786,7 +787,7 @@ static void server_main_loop(int sock, char **device_name,
 				 * Start the proxy work in the new socket.
 				 */
 #endif
-				serve_client(new_sock,device_name, n_dev, 0);
+				serve_client(new_sock, device_name, n_dev, 0);
 				exit(0);
 #if DEBUG == 0
 		}
@@ -839,7 +840,7 @@ int main (int argc, char** argv)
 	char*			username = strdup("nobody");
 	int			sock;
 
-	char **device_name = NULL; 
+	const char *const* device_name = NULL; 
 	const char *floppy0 = "/dev/fd0";
 	unsigned int n_dev;
 
@@ -883,10 +884,10 @@ int main (int argc, char** argv)
 		}
 
 	if(optind < argc) {
-		device_name = argv + optind;
-		n_dev = argc - optind;
+		device_name = (const char * const *) argv + optind;
+		n_dev = (unsigned int) (argc - optind);
 	} else {
-		device_name = (char **)&floppy0;
+		device_name = &floppy0;
 		n_dev = 1;
 	}
 
@@ -1035,7 +1036,7 @@ static void send_reply64(int rval, io_buffer sock, mt_off_t len) {
 	Packet reply = newPacket();
 
 	make_new(reply, 12);
-	put_qword(reply, 0, len);
+	put_qword(reply, 0, (Qword) len);
 	if (rval == -1) {
 		put_dword(reply, 8, 0);
 	} else {
@@ -1053,8 +1054,8 @@ static void cleanup(int x UNUSEDP) {
 
 #include "lockdev.h"
 
-void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
-		  int close_stderr) {
+void serve_client(int sockhandle, const char *const*device_name,
+		  unsigned int n_dev, int close_stderr) {
 	Packet opcode;
 	Packet parm;
 
@@ -1227,7 +1228,7 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 				} else {
 					rval = write_packet(parm, devFd);
 				}
-				send_reply(devFd, sock, rval);
+				send_reply(devFd, sock, (Dword) rval);
 				break;
 			case OP_SEEK:
 #if DEBUG
@@ -1235,7 +1236,8 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 #endif
 
 				lseek(devFd, 
-				      get_dword(parm, 0), get_dword(parm, 4));
+				      get_dword(parm, 0),
+				      (int) get_dword(parm, 4));
 				send_reply(devFd, 
 					   sock, 
 					   (Dword) lseek(devFd, 0, SEEK_CUR));
@@ -1253,7 +1255,8 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 				fprintf(stderr, "SEEK64:\n");
 #endif
 				mt_lseek(devFd, 
-					 get_qword(parm,0), get_dword(parm,8));
+					 (mt_off_t) get_qword(parm,0),
+					 (int) get_dword(parm,8));
 				send_reply64(devFd, 
 					     sock, 
 					     mt_lseek(devFd, 0, SEEK_CUR));
