@@ -25,18 +25,16 @@
 #include "stream.h"
 #include "mtools.h"
 #include "fsP.h"
-#include "plain_io.h"
-#include "floppyd_io.h"
-#include "xdf_io.h"
 #include "buffer.h"
 #include "file_name.h"
+#include "open_image.h"
 
 #define FULL_CYL
 
 /*
  * Read the boot sector.  We glean the disk parameters from this sector.
  */
-static int read_boot(Stream_t *Stream, union bootsector * boot, int size)
+static int read_boot(Stream_t *Stream, union bootsector * boot, size_t size)
 {
 	size_t boot_sector_size; /* sector size, as stored in boot sector */
 
@@ -46,7 +44,7 @@ static int read_boot(Stream_t *Stream, union bootsector * boot, int size)
 	if(size > MAX_BOOT)
 		size = MAX_BOOT;
 
-	if (force_read(Stream, boot->characters, 0, size) != size)
+	if (force_read(Stream, boot->characters, 0, size) != (ssize_t) size)
 		return -1;
 
 	boot_sector_size = WORD(secsiz);		
@@ -116,8 +114,218 @@ Stream_t *GetFs(Stream_t *Fs)
 	return Fs;
 }
 
+static void boot_to_geom(struct device *dev, int media,
+			 union bootsector *boot) {
+	uint32_t tot_sectors;
+	int BootP, Infp0, InfpX, InfTm;
+	int j;
+	unsigned char sum;
+	uint16_t sect_per_track;
+	struct label_blk_t *labelBlock;
+
+	dev->ssize = 2; /* allow for init_geom to change it */
+	dev->use_2m = 0x80; /* disable 2m mode to begin */
+
+	if(media == 0xf0 || media >= 0x100){		
+		dev->heads = WORD(nheads);
+		dev->sectors = WORD(nsect);
+		tot_sectors = DWORD(bigsect);
+		SET_INT(tot_sectors, WORD(psect));
+		sect_per_track = dev->heads * dev->sectors;
+		if(sect_per_track == 0) {
+		    if(mtools_skip_check) {
+			/* add some fake values if sect_per_track is
+			 * zero. Indeed, some atari disks lack the
+			 * geometry values (i.e. have zeroes in their
+			 * place). In order to avoid division by zero
+			 * errors later on, plug 1 everywhere
+			 */
+			dev->heads = 1;
+			dev->sectors = 1;
+			sect_per_track = 1;
+		    } else {
+			fprintf(stderr, "The devil is in the details: zero number of heads or sectors\n");
+			exit(1);
+		    }
+		}
+		dev->tracks = tot_sectors / sect_per_track;
+		if(tot_sectors % sect_per_track)
+			/* round size up */
+			dev->tracks++;
+		
+		BootP = WORD(ext.old.BootP);
+		Infp0 = WORD(ext.old.Infp0);
+		InfpX = WORD(ext.old.InfpX);
+		InfTm = WORD(ext.old.InfTm);
+		
+		if(WORD(fatlen)) {
+			labelBlock = &boot->boot.ext.old.labelBlock;
+		} else {
+			labelBlock = &boot->boot.ext.fat32.labelBlock;
+		}
+
+		if (boot->boot.descr >= 0xf0 &&
+		    has_BPB4 &&
+		    strncmp( boot->boot.banner,"2M", 2 ) == 0 &&
+		    BootP < 512 && Infp0 < 512 && InfpX < 512 && InfTm < 512 &&
+		    BootP >= InfTm + 2 && InfTm >= InfpX && InfpX >= Infp0 && 
+		    Infp0 >= 76 ){
+			for (sum=0, j=63; j < BootP; j++) 
+				sum += boot->bytes[j];/* checksum */
+			dev->ssize = boot->bytes[InfTm];
+			if (!sum && dev->ssize <= 7){
+				dev->use_2m = 0xff;
+				dev->ssize |= 0x80; /* is set */
+			}
+		}
+		dev->sector_size = WORD(secsiz);
+	} else
+		if(setDeviceFromOldDos(media, dev) < 0)
+			exit(1);
+}
+
 /**
- * Tries out all device definitions for the given drive number, until one
+ * Tries out one device definition for the given drive number
+ * Parameters
+ *  - dev: device definition to try
+ *  - mode: file open mode
+ *  - out_dev: device parameters (geometry, etc.) are returned here
+ *  - boot: boot sector is read from the disk into this structure
+ *  - name: "name" of device definition (returned)
+ *  - media: media byte is returned here (ored with 0x100 if there is a
+ *    BIOS Parameter block present)
+ *  - maxSize: maximal size supported by (physical) drive returned here
+ *  - try_writable: whether to try opening it writable from the get-go,
+ *     even if not specified as writable in mode (used for mlabel)
+ *  - isRop: whether device is read-only is returned here
+ * Return value:
+ *  - a Stream allowing to read from this device, must be closed by caller
+ *
+ * If a geometry change is needed, drive is re-opened RW, as geometry
+ * change ioctl needs write access. However, in such case, the lock
+ * acquired is still only a read lock.
+ */
+static Stream_t *try_device(struct device *dev,
+			    int mode, struct device *out_dev,
+			    union bootsector *boot,
+			    char *name, int *media, mt_size_t *maxSize,
+			    int *isRop, int try_writable,
+			    char *errmsg)
+{
+	int retry_write;
+	int have_read_bootsector=0;
+	int modeFlags = mode & ~O_ACCMODE;
+	int openMode;
+	int lockMode;
+
+	*out_dev = *dev;
+	expand(dev->name,name);
+#ifdef USING_NEW_VOLD
+	strcpy(name, getVoldName(dev, name));
+#endif
+
+	if(try_writable) {
+		/* Caller asks up to try first read-write, and only fall back
+		 * if not feasible */
+		openMode = O_RDWR | modeFlags;
+	} else {
+		openMode = mode;
+	}
+	lockMode = openMode;
+
+	for(retry_write=0; retry_write<2; retry_write++) {
+		Stream_t *Stream;
+		int r;
+		int geomFailure=0;
+
+		if(retry_write)
+			mode |= O_RDWR;
+
+		Stream = OpenImage(out_dev, dev, name, openMode, errmsg,
+				   0, lockMode,
+				   maxSize, &geomFailure, NULL);
+		if(Stream == NULL) {
+			if(geomFailure && (mode & O_ACCMODE) == O_RDONLY) {
+				/* Our first attempt was to open read-only,
+				   but this resulted in failure setting the 
+				   geometry */
+				openMode = modeFlags | O_RDWR;
+				continue;
+			}
+
+			if(try_writable &&
+			   (errno == EPERM ||
+			    errno == EACCES ||
+			    errno == EROFS)) {
+				/* Our first attempt was to open
+				 * read-write, but this resulted in a
+				 * read-protection problem */
+				lockMode = openMode = modeFlags | O_RDONLY;
+				continue;
+			}
+			return NULL;
+		}
+		if(!have_read_bootsector) {
+			/* read the boot sector */
+			if ((r=read_boot(Stream, boot, out_dev->blocksize)) < 0){
+				sprintf(errmsg,
+					"init %c: could not read boot sector",
+					dev->drive);
+				FREE(&Stream);
+				return NULL;
+			}
+
+			if((*media= get_media_type(Stream, boot)) <= 0xf0 ){
+				if (boot->boot.jump[2]=='L')
+					sprintf(errmsg,
+						"diskette %c: is Linux LILO, not DOS",
+						dev->drive);
+				else
+					sprintf(errmsg,"init %c: non DOS media", dev->drive);
+				FREE(&Stream);
+				return NULL;
+			}
+			have_read_bootsector=1;
+		}
+
+		/* set new parameters, if needed */
+		errno = 0;
+		boot_to_geom(out_dev, *media, boot);
+		if(SET_GEOM(Stream, out_dev, dev)){
+			if(errno == EBADF || errno == EPERM) {
+				/* Retry with write */
+				FREE(&Stream);
+				openMode = modeFlags | O_RDWR;
+				continue;
+			}
+			if(errno)
+#ifdef HAVE_SNPRINTF
+				snprintf(errmsg, 199,
+					 "Can't set disk parameters for %c: %s",
+					 dev->drive, strerror(errno));
+#else
+			sprintf(errmsg,
+				"Can't set disk parameters for %c: %s",
+				drive, strerror(errno));
+#endif
+			else
+				sprintf(errmsg,
+					"Can't set disk parameters for %c",
+					dev->drive);
+			FREE(&Stream);
+			return NULL;
+		}
+		if(isRop) {
+			*isRop = (openMode & O_ACCMODE) == O_RDONLY;
+		}
+		return Stream;
+	}
+	return NULL;
+}
+
+
+/**
+ * Tries out all device definitions for the given drive letter, until one
  * is found that is able to read from the device
  * Parameters
  *  - drive: drive letter to check
@@ -138,124 +346,43 @@ Stream_t *find_device(char drive, int mode, struct device *out_dev,
 		      int *isRop)
 {
 	char errmsg[200];
-	Stream_t *Stream;
 	struct device *dev;
-	int r;
-	int isRo=0;
-
-	Stream = NULL;
+	
 	sprintf(errmsg, "Drive '%c:' not supported", drive);
 					/* open the device */
 	for (dev=devices; dev->name; dev++) {
-		FREE(&Stream);
+		Stream_t *Stream;
+		int isRo;
+		isRo=0;
 		if (dev->drive != drive)
 			continue;
-		*out_dev = *dev;
-		expand(dev->name,name);
-#ifdef USING_NEW_VOLD
-		strcpy(name, getVoldName(dev, name));
-#endif
 
-		Stream = 0;
-		if(out_dev->misc_flags & FLOPPYD_FLAG) {
-		    Stream = 0;
-#ifdef USE_FLOPPYD
-		    Stream = FloppydOpen(out_dev, name, mode,
-					 errmsg, maxSize);
-#endif
-		} else {
-
-#ifdef USE_XDF
-		    Stream = XdfOpen(out_dev, name, mode, errmsg, 0);
-		    if(Stream) {
-			out_dev->use_2m = 0x7f;
-			if(maxSize)
-			    *maxSize = max_off_t_31;
-		    }
-#endif
-
-
-		    if (!Stream)
-			Stream = SimpleFileOpen(out_dev, dev, name,
-						isRop ? mode | O_RDWR: mode,
-						errmsg, 0, 1, maxSize);
-
-		    if(Stream) {
-			isRo=0;
-		    } else if(isRop &&
-		       (errno == EPERM || errno == EACCES || errno == EROFS)) {
-			Stream = SimpleFileOpen(out_dev, dev, name,
-						mode | O_RDONLY,
-						errmsg, 0, 1, maxSize);
-			if(Stream) {
-				isRo=1;
-			}
-		    }
+		Stream = try_device(dev, mode, out_dev,
+				    boot,
+				    name, media, maxSize,
+				    &isRo, isRop != NULL,
+				    errmsg);
+		if(Stream) {
+			if(isRop)
+				*isRop = isRo;
+			return Stream;
 		}
-
-		if( !Stream)
-		    continue;
-
-		/* read the boot sector */
-		if ((r=read_boot(Stream, boot, out_dev->blocksize)) < 0){
-			sprintf(errmsg,
-				"init %c: could not read boot sector",
-				drive);
-			continue;
-		}
-
-		if((*media= get_media_type(Stream, boot)) <= 0xf0 ){
-			if (boot->boot.jump[2]=='L')
-				sprintf(errmsg,
-					"diskette %c: is Linux LILO, not DOS",
-					drive);
-			else
-				sprintf(errmsg,"init %c: non DOS media", drive);
-			continue;
-		}
-
-		/* set new parameters, if needed */
-		errno = 0;
-		if(SET_GEOM(Stream, out_dev, dev, *media, boot)){
-			if(errno)
-#ifdef HAVE_SNPRINTF
-				snprintf(errmsg, 199,
-					"Can't set disk parameters for %c: %s",
-					drive, strerror(errno));
-#else
-				sprintf(errmsg,
-					"Can't set disk parameters for %c: %s",
-					drive, strerror(errno));
-#endif
-			else
-				sprintf(errmsg,
-					"Can't set disk parameters for %c",
-					drive);
-			continue;
-		}
-		break;
 	}
 
 	/* print error msg if needed */
-	if ( dev->drive == 0 ){
-		FREE(&Stream);
-		fprintf(stderr,"%s\n",errmsg);
-		return NULL;
-	}
-	if(isRop)
-		*isRop = isRo;
-	return Stream;
+	fprintf(stderr,"%s\n",errmsg);
+	return NULL;
 }
 
 
 Stream_t *fs_init(char drive, int mode, int *isRop)
 {
-	int blocksize;
+	uint32_t blocksize;
 	int media;
-	int disk_size = 0;	/* In case we don't happen to set this below */
-	size_t tot_sectors;
+	size_t disk_size = 0;	/* In case we don't happen to set this below */
+	uint32_t tot_sectors;
 	char name[EXPAND_BUF];
-	int cylinder_size;
+	unsigned int cylinder_size;
 	struct device dev;
 	mt_size_t maxSize;
 
@@ -291,8 +418,10 @@ Stream_t *fs_init(char drive, int mode, int *isRop)
 		 * else result of the AND would be 0x1xx
 		 */
 		struct OldDos_t *params=getOldDosByMedia(media);
-		if(params == NULL)
+		if(params == NULL) {
+			fprintf(stderr, "Unknown media byte %02x\n", media);
 			return NULL;
+		}
 		This->cluster_size = params->cluster_size;
 		tot_sectors = cylinder_size * params->tracks;
 		This->fat_start = 1;
@@ -305,7 +434,7 @@ Stream_t *fs_init(char drive, int mode, int *isRop)
 		This->fat_bits = 12;
 	} else {
 		struct label_blk_t *labelBlock;
-		int i;
+		unsigned int i;
 		
 		This->sector_size = WORD_S(secsiz);
 		if(This->sector_size > MAX_SECTOR){
@@ -402,6 +531,7 @@ Stream_t *fs_init(char drive, int mode, int *isRop)
 
 	/* read the FAT sectors */
 	if(fat_read(This, &boot, tot_sectors, dev.use_2m&0x7f)){
+		fprintf(stderr, "Error reading FAT\n");
 		This->num_fat = 1;
 		FREE(&This->Next);
 		Free(This->Next);
@@ -411,6 +541,7 @@ Stream_t *fs_init(char drive, int mode, int *isRop)
 	/* Set the codepage */
 	This->cp = cp_open(dev.codepage);
 	if(This->cp == NULL) {
+		fprintf(stderr, "Error setting code page\n");
 		fs_free((Stream_t *)This);
 		FREE(&This->Next);
 		Free(This->Next);
@@ -430,7 +561,15 @@ char getDrive(Stream_t *Stream)
 		return This->drive;
 }
 
-int fsPreallocateClusters(Fs_t *Fs, long size)
+/*
+ * Upper layer asks to pre-allocated more additional clusters
+ * Parameters:
+ *   size: new additional clusters to pre-allocate
+ * Return:
+ *   0  if pre-allocation was granted
+ *  -1  if not enough clusters could be found
+ */
+int fsPreallocateClusters(Fs_t *Fs, uint32_t size)
 {
 	if(size > 0 && getfreeMinClusters((Stream_t *)Fs, size) != 1)
 		return -1;
@@ -438,3 +577,16 @@ int fsPreallocateClusters(Fs_t *Fs, long size)
 	Fs->preallocatedClusters += size;
 	return 0;
 }
+
+/*
+ * Upper layer wants to release some clusters that it had
+ * pre-allocated before Usually done because they have now been really
+ * allocated, and thus pre-allocation needs to be released to prevent
+ * counting them twice.
+ * Parameters:
+ *   size: new additional clusters to pre-allocate
+ */
+void fsReleasePreallocateClusters(Fs_t *Fs, uint32_t size)
+{
+	Fs->preallocatedClusters -= size;
+}
